// IDSQLDirect.h: interface for the CIDSQLDirect class.
//
// The main use of this class is to connect to an ODBC data source
// and issue SQL commands.  This class uses STL mechanisms
// instead of the CPtrArray MFC class.
//
// Author: Ian Duff - 30 December 2001
// Email : ianduff@bold.net.au
//
// Many thanks to Dave Merner for his article on direct SQL calls with ODBC
//////////////////////////////////////////////////////////////////////

#if !defined(   __IDSQLDIRECT_H__ )
#define         __IDSQLDIRECT_H__ 

// ---------- Module includes

#include "IDSQLColumn.h"
#include <vector>       // We'll store a vector of columns
#include <algorithm>    // find, etc
#include <functional>   // STL function access
#include <fstream>		// ifstream, etc
#include <ctype.h>      // toupper macros, etc

// ---------- Module namespaces 

using namespace std;

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

/******************************************************************************
* Class CIDSQLDirect
*
* This class encompasses most of the required functionality 
* for issuing SQL commands through an ODBC connection.
*
* The CIDSQLColumn operators will now allow you to use other STL
* mechanisms like sort:
* sort(m_vColArray.begin(), m_vColArray.end() );
******************************************************************************/

class CIDSQLDirect  
{
public:

    /* Default constructor */
	CIDSQLDirect();

    /* Default destructor */
	virtual ~CIDSQLDirect();

    /* Add a column (CIDSQLColumn) to the vector of columns */
	void            AddColumn( int, string );

    /* Free resources and clear the vector of columns */
	CIDSQLDirect&   Close();

    /* Issue a database commit through this class */
	bool            Commit();

    /* Connect to the database */
	int             Connect( LPCSTR );

    /* Connect to the database using the supplied DSN */
	int             ConnectDSN( LPCSTR );

    /* Connect to the database using the supplied DSN with an STL string */
    int             ConnectDSN( string s ) { return( ConnectDSN( s.c_str() ) ); }

    /* Calls AfxMessageBox with a constructed error message */
	void            DisplayError();

    /* Execute the incoming SQL query */
	int             ExecuteSQL( LPCSTR);

    /* Execute the query contained in the incoming file name */
	int				ExecuteQueryFile( string );

    /* Fetch the next row */
	int             Fetch();

    /* Find the incoming column number */
	int             FindColumn( int );

    /* Get the column indicated by the incoming number */
	string          GetCol( int );
    
    /* Get the column indicated by the incoming string */
	string          GetCol( string, string = "" );

    /* Get a pointer to the column indicated by the incoming number */
	CIDSQLColumn*   GetColumn( int );

    /* Get a pointer to the column indicated by the incoming strings
     * where the first is the named column and the second is the table name.
    */
	int             GetColumnID( string, string );

    /* Get the name of the column indicated by the incoming number */
	string          GetColumnName( int );

    /* Get the type of the column indicated by the incoming number */
	int             GetColumnType( int );

    /* Get the error string */
	string          GetError();

    /* Get the number of columns returned by a successful query */
	int             GetNumColumns();

    /* Print all of the column names */
    CIDSQLDirect&   PrintColumnNames();

    /* Rollback transaction command */
	bool            Rollback();
    
    /* Set the internal name of the connection DSN */
    CIDSQLDirect&   SetDSN( LPCSTR );

    /* Set the password name */
	CIDSQLDirect&   SetPassword( LPCSTR );

    /* Set the user name */
	CIDSQLDirect&   SetUserName( LPCSTR );

    /* Suppress/no spression of errors */
	inline void CIDSQLDirect::SupressErrors( bool bSupress = true ) 
    {  m_bSupressErrors = bSupress; }

    /* Utility function trims blank spaces from the right of a string */
	string          TrimRight( LPCSTR );

private:
    /* Initialise all internal members to a ready state */
	void            Init();

	SQLRETURN       m_nReturn;                  // Internal SQL error code
    string          m_strSQL;                   // SQL string
    string          m_strErrorMsg;	            // SQL error message
	HENV            m_hEnv;                     // Handle to environment
	HDBC            m_hDBC;                     // Handle to database connection
	HSTMT           m_hStmt;                    // Handle to Statement

    vector< CIDSQLColumn > m_vColArray;         // vector of columns
    vector< CIDSQLColumn >::iterator m_colIter; // column iterator

	bool            m_bSupressErrors;           // Ignore errors if set to true
    string          m_strPassword;              // User password
    string          m_strUserName;              // User name
    string          m_strDsn;                   // DSN connection string
};

#endif // !defined( __IDSQLDIRECT_H__ )
